/*
 This software is Copyright (c) 2007-2008
 Adam Maxwell. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 - Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 
 - Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in
 the documentation and/or other materials provided with the
 distribution.
 
 - Neither the name of Adam Maxwell nor the names of any
 contributors may be used to endorse or promote products derived
 from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import <Cocoa/Cocoa.h>

static void writeError(NSString *msg)
{
    msg = [NSString stringWithFormat:@"@article{error,\ntitle = {*** ERROR *** %@}}", msg];
    [(NSFileHandle *)[NSFileHandle fileHandleWithStandardOutput] writeData:[msg dataUsingEncoding:NSUTF8StringEncoding]];
}

// this is a sample function to clean PubMed-style authors
static NSString *pubmedToBibTeXValue(NSString *value, NSString *name)
{
    NSString *stringValue = value;
    
    if ([name caseInsensitiveCompare:@"author"] == NSOrderedSame) {
        
        // this sucks totally...didn't someone realize that it would be helpful to provide this in a parseable format?
        // we get:  "Freely IP, Away GO, Weiser BUD"
        // we need: "Freely, I. P. and Away, G. O. and Weiser, B. U. D."
        NSArray *authors = [stringValue componentsSeparatedByString:@", "];
        
        // what happens with corporate authors?
        NSMutableArray *fixedAuthors = [NSMutableArray arrayWithCapacity:[authors count]];
        NSEnumerator *e = [authors objectEnumerator];
        NSString *name;
        while (name = [e nextObject]) {
            // so, now we have @"Freely IP"
            NSRange r = [name rangeOfCharacterFromSet:[NSCharacterSet whitespaceCharacterSet]];
            NSString *last;
            if (r.length) {
                last = [name substringWithRange:NSMakeRange(0, r.location)];
                last = [last stringByAppendingString:@", "];
                
                NSRange firstRange = NSMakeRange(r.location + 1, [name length] - r.location - 1);
                while (firstRange.length) {
                    last = [last stringByAppendingFormat:@"%@. ", [name substringWithRange:NSMakeRange(firstRange.location, 1)]];
                    firstRange.location += 1;
                    firstRange.length -= 1;
                }
                [fixedAuthors addObject:last];
                
            }else {
                // assume the entire string is a last name
                [fixedAuthors addObject:name];
            }
        }
        stringValue = [fixedAuthors componentsJoinedByString:@" and "];

    }
    
    return stringValue;
}

int main (int argc, const char * argv[]) {
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];

    NSArray *args = [[NSProcessInfo processInfo] arguments];
    unsigned i, iMax = [args count];
    BOOL isPubmed = NO;
    
    @try{
        for (i = 1; i < iMax; i++) {
            
            if ([[args objectAtIndex:i] caseInsensitiveCompare:@"--pubmed"]) {
                isPubmed = YES;
                continue;
            }
            
            NSURL *theURL = [NSURL URLWithString:[args objectAtIndex:i]];
            
            // we cleverly write errors to stdout as BibTeX and lie about our return status so BibDesk displays them
            if (nil == theURL) {
                writeError([NSString stringWithFormat:@"Invalid URL %@", [args objectAtIndex:i]]);
                return EXIT_SUCCESS;
            }
            
            NSError *error;
            NSXMLDocument *xmlDoc = [[NSXMLDocument alloc] initWithContentsOfURL:theURL options:NSXMLNodeOptionsNone error:&error];
            
            if (nil == xmlDoc) {
                writeError([NSString stringWithFormat:@"Invalid XML document at %@", [theURL absoluteString]]);
                writeError([NSString stringWithFormat:@"NSXMLDocument returned error %@", [error description]]);
                return EXIT_SUCCESS;
            }
                    
            // get an array of all the item nodes
            NSArray *items = [xmlDoc nodesForXPath:@"/rss[1]/channel[1]/item" error:&error];
            
            if (nil == items) {
                writeError([NSString stringWithFormat:@"XPath /rss[1]/channel[1]/item returned error %@", [error description]]);
                return EXIT_SUCCESS;
            }
            
            // get the first node and then walk the tree
            NSXMLNode *node = [items count] ? [items objectAtIndex:0] : nil;
            NSMutableString *btString = [NSMutableString string];
            NSCharacterSet *whitespaceSet = [NSCharacterSet whitespaceCharacterSet];
            NSFileHandle *fh = [NSFileHandle fileHandleWithStandardOutput];
            
            while (nil != node) {
                
                // now walk each item's tree, instead of hard-coding specific paths
                NSXMLNode *child = [node childCount] ? [node childAtIndex:0] : nil;
                [btString setString:@"@article{FixMe,\n"];

                while (nil != child) {
                                    
                    NSString *name = [child name];
                    NSString *stringValue = [[child stringValue] stringByTrimmingCharactersInSet:whitespaceSet];
                    
                    // special-case so BibDesk recognizes this as a URL
                    if ([name caseInsensitiveCompare:@"link"] == NSOrderedSame) {
                        name = @"url";
                    } else if ([name caseInsensitiveCompare:@"description"] == NSOrderedSame) {
                        // try to read this as HTML
                        NSAttributedString *attrString = [[NSAttributedString alloc] initWithHTML:[stringValue dataUsingEncoding:NSUTF8StringEncoding] documentAttributes:NULL];
                        if (attrString) {
                            stringValue = [attrString string];
                            [attrString autorelease];
                        }
                    }
                    
                    // We may clean pubmed style authors. Here you could add any additional cleaning and converting as necessary
                    if (isPubmed)
                        stringValue = pubmedToBibTeXValue(stringValue, name);
                    
                    if (stringValue)
                        [btString appendFormat:@"%@ = {%@}", name, [stringValue stringByTrimmingCharactersInSet:whitespaceSet]];
                    
                    // get the next field-value pair
                    child = [child nextSibling];
                    
                    // not really necessary, but nicer to look at
                    if (child)
                        [btString appendFormat:@",\n"];
                }

                [btString appendString:@"\n}\n\n"];
                
                // write this item to stdout as BibTeX
                [fh writeData:[btString dataUsingEncoding:NSUTF8StringEncoding]];
                
                // see if we have another item node
                node = [node nextSibling];
            }
            
            [xmlDoc release];
        }
    }
    @catch(id exception) {
        writeError([NSString stringWithFormat:@"Caught exception \"%@\" while reading feed"]);
    }

    [pool release];
    return EXIT_SUCCESS;
}
